/* ReSCAL - Engine entry
 *
 * Copyright (C) 2011
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <locale.h>

#include "defs.h"
#include "macros_public.h"
#include "rescal.h"
#include "monitor.h"
#ifdef PARALLEL
#include "synchro.h"
#endif
#include "cellules.h"
#include "simul.h"
#include "terre.h"
#include "doublets.h"
#include "transitions.h"
#include "param.h"
#include "trace.h"
#include "surface.h"

#ifdef COLLISIONS
#include "collisions.h"
#endif

extern unsigned char opt_info;
extern unsigned long int iter;        // nombre d'iterations
extern double temps;                    // temps reel simule
#ifdef PARALLEL
extern int mode_par;    //mode parallele
#endif

int etape_simulation = SIMUL_DEBUT; //etape courante de la simulation
int stop_simul = 0;                 //semaphore (pour le multithreading)
int fin_simul = 0;                  //semaphore (pour l'arret des callbacks)

void rescal_usage()
{
  printf("ReSCAL - %s model\n",MOD_NAME);
  printf("version %s (%s)\n",VER_NUM,VER_DAT);
#ifdef __x86_64
  printf("compiled for x86_64\n");
  /*
  printf("size of long: %ld\n", sizeof(long));
  printf("size of int: %ld\n", sizeof(int));
  printf("size of address: %ld\n", sizeof(char*));
  */
#endif

/*#ifdef DUMP_IMAGE
  printf("use of libgd\n");
#endif*/

#ifdef COLLISIONS
  printf("lattice gaz enabled\n");
#endif

  printf("\nusage: rescal PARAMETERS_FILE [OPTIONS]\n\n");
}


void rescal_init(int argc, char ** argv)
{
  LogPrintf("rescal is running ...\n");

#ifdef NUM_MODE
  LogPrintf("format des valeurs numeriques : %s \n", NUM_MODE);
  setlocale(LC_NUMERIC, NUM_MODE);
#endif

  lecture_parametres(argv[1]);
  init_signal();

  etape_simulation = SIMUL_INIT;

#ifdef PARALLEL
  if (mode_par) synchro_init(argc, argv);
#endif

  cree_terre();
  calcule_couloir();
  calcule_h_plaf();
#ifdef PARALLEL
  cree_tunnels();
#endif

  init_Ncel();

  init_transitions();
  init_db_pos();

#ifdef COLLISIONS
  init_collisions();
#endif

  if (opt_info){
    dump_cellules();
    dump_doublets();
    dump_transitions();
  }

#ifdef MODELE_DUN
  calcule_alti(GR, ALTI_MODE_BAS);
  calcule_normales();
#endif

#if defined(MODELE_DUN) && defined(COLLISIONS)
  //dump_cgv();
#endif

#ifdef SURRECTIONS
  //regul_niveau();
#endif

#if MODE_TRACE
  trace_init();
#endif

  etape_simulation = SIMUL_PAUSE;

  LogPrintf("fin rescal_init()\n");

}

int rescal(int n)
{
  static unsigned char start=0;
  unsigned char res=0;

  if (!start) {
#ifdef PARALLEL
    if (mode_par) synchro_start();
#endif
    //monitor_signal();
    LogPrintf("debut de la simulation ...\n");
    start = 1;
  }

  res = 1;
  while ((!stop_simul) && (res)){
    etape_simulation = SIMUL_EN_COURS;
#if defined(TRACE_SRC) || defined(TRACE_AIRE) || defined(TRACE_PAR)
    res = trace_init_loop(0);
    if (!res) break;
#endif
    res = simul_terre(n, &stop_simul);
#if defined(TRACE_AIRE) || defined(TRACE_PAR)
    trace_end_loop();
    res = 1;
#endif
  }

  if (!res){
    LogPrintf("fin de la simulation\n");
    etape_simulation = SIMUL_FIN;
    //stop_dump();
    fin_simul = 1;
    //monitor_signal();
  }
  else
    etape_simulation = SIMUL_PAUSE;

  return (res);
}

/*
void rescal_stop()
{
  if  ((etape_simulation == SIMUL_INIT) ||  (etape_simulation == SIMUL_EN_COURS)){
    stop_simul = 1;
    LogPrintf("arret de la simulation !\n");
    while ((etape_simulation == SIMUL_INIT) ||  (etape_simulation == SIMUL_EN_COURS));  //boucle d'attente
  }
}

void rescal_quit()
{
  rescal_stop();
#ifdef TRACE_TRANS
  trace_trans_quit();
#endif
#ifdef TRACE3D_CEL
  trace3d_cel_quit();
#endif
  fin_db_pos();
  apocalypse();
}
*/

